import sys
from fnmatch import fnmatch
from typing import List

import typer
from ordered_set import OrderedSet
from pynmrstar import Entry
from typer import Argument, Option

from nef_pipelines.lib.util import chunks, get_pipe_file
from nef_pipelines.tools.chains import chains_app

app = typer.Typer()

# TODO: it would be nice to put the chains with the first molecular system frame


# noinspection PyUnusedLocal
@chains_app.command()
def rename(
    old: str = Argument(..., help="old chain code"),
    new: str = Argument(..., help="new chain code"),
    comment: bool = Option(False, "--comment", help="prepend comment to chains"),
    verbose: bool = Option(False, "-v", "--verbose", help="print verbose info"),
    use_category: bool = Option(
        False,
        "-c",
        "--category",
        help="select frames to rename chains in by category rather than name",
    ),
    frames: List[str] = Option(
        [],
        "-f",
        "--frame",
        help="limit changes to a a particular frame by name [or category if --category is set], note: wildcards are "
        "allowed, repeated uses add more frames",
    ),
):
    """- change the name of chains across one or multuiple frames"""

    lines = "".join(get_pipe_file([]).readlines())
    entry = Entry.from_string(lines)

    changes = 0
    changed_frames = OrderedSet()
    for save_frame in entry:
        process_frame = True
        for frame_selector in frames:

            if use_category:
                if not fnmatch(save_frame.category, f"*{frame_selector}*"):

                    process_frame = False
            else:
                if not fnmatch(save_frame.name, f"*{frame_selector}*"):
                    process_frame = False

        if process_frame:
            for loop in save_frame.loop_iterator():
                for tag in loop.get_tag_names():
                    tag_parts = tag.split(".")
                    if tag_parts[-1].startswith("chain_code"):
                        tag_values = loop[tag]
                        for i, row in enumerate(tag_values):
                            if row == old:
                                tag_values[i] = new
                                changes += 1
                                changed_frames.add(save_frame.name)

                        loop[tag] = tag_values

    if verbose:
        comment = "# " if comment else ""
        out = sys.stderr if not comment else sys.stdout
        if changes >= 1:
            print(
                f"{comment}rename chain: {changes} changes made in the following frames",
                file=out,
            )
            for chunk in chunks(changed_frames, 5):
                print(f'{comment}  {", ".join(chunk)}', file=out)

        else:
            print(f"{comment}rename chain: no changes made", file=out)
        print(file=out)

    print(entry)



from itertools import islice
from textwrap import dedent

import pytest

from lib.sequence_lib import translate_1_to_3, sequence_3let_to_sequence_residues, BadResidue, frame_to_chains, \
    count_residues, get_chain_starts, offset_chain_residues, chain_code_iter
from lib.structures import SequenceResidue
from pynmrstar import Saveframe

ABC_SEQUENCE_1LET = 'acdefghiklmnpqrstvwy'
ABC_SEQUENCE_3LET = (
        'ALA',
        'CYS',
        'ASP',
        'GLU',
        'PHE',
        'GLY',
        'HIS',
        'ILE',
        'LYS',
        'LEU',
        'MET',
        'ASN',
        'PRO',
        'GLN',
        'ARG',
        'SER',
        'THR',
        'VAL',
        'TRP',
        'TYR'
)

ABC_SEQUENCE_RESIDUES = [SequenceResidue('A', i, residue) for (i, residue) in enumerate(ABC_SEQUENCE_3LET, start=1)]


def test_1let_3let():

    assert len(ABC_SEQUENCE_1LET) == 20
    assert len(ABC_SEQUENCE_3LET) == 20

    assert list(ABC_SEQUENCE_3LET) == translate_1_to_3(ABC_SEQUENCE_1LET)


def test_bad_1let_3let():
    BAD_SEQUENCE = 'acdefghiklmonpqrstvwy'

    msgs = '''\
              unknown residue O
              at residue 12
              sequence: acdefghiklmonpqrstvwy
              ^
              '''

    msgs = dedent(msgs)
    msgs = msgs.split('\n')

    with pytest.raises(BadResidue) as exc_info:
        translate_1_to_3(BAD_SEQUENCE)

    for msg in msgs:
        assert msg in exc_info.value.args[0]


def test_3let_sequence_residue():

    sequence_residues = sequence_3let_to_sequence_residues(ABC_SEQUENCE_3LET)

    assert sequence_residues == ABC_SEQUENCE_RESIDUES


def test_3let_sequence_residue_diff_chain():
    sequence_residues = sequence_3let_to_sequence_residues(ABC_SEQUENCE_3LET, chain_code='B')

    expected = [SequenceResidue('B', residue.sequence_code, residue.residue_name) for residue in ABC_SEQUENCE_RESIDUES]

    assert sequence_residues == expected


TEST_DATA_MULTI_CHAIN = """
    save_nef_molecular_system
       _nef_molecular_system.sf_category   nef_molecular_system
       _nef_molecular_system.sf_framecode  nef_molecular_system

       loop_
          _nef_sequence.index
          _nef_sequence.chain_code
          _nef_sequence.sequence_code
          _nef_sequence.residue_name
          _nef_sequence.linking
          _nef_sequence.residue_variant
          _nef_sequence.cis_peptide
          _nef_sequence.ccpn_comment
          _nef_sequence.ccpn_chain_role
          _nef_sequence.ccpn_compound_name
          _nef_sequence.ccpn_chain_comment

         1    A   3    HIS   .   .   .   .   .   Sec5   .    
         2    A   4    MET   .   .   .   .   .   Sec5   .    
         3    B   5    ARG   .   .   .   .   .   Sec5   .    
         4    B   6    GLN   .   .   .   .   .   Sec5   .    
         5    C   7    PRO   .   .   .   .   .   Sec5   .       

       stop_

    save_

    """

def test_list_chains():

    test_frame = Saveframe.from_string(TEST_DATA_MULTI_CHAIN)
    chains = frame_to_chains(test_frame)

    assert chains == list(['A', 'B', 'C'])


def test_list_chains_no_chains():
    TEST_DATA = """
    save_nef_molecular_system
       _nef_molecular_system.sf_category   nef_molecular_system
       _nef_molecular_system.sf_framecode  nef_molecular_system

       loop_
          _nef_sequence.index
          _nef_sequence.chain_code
          _nef_sequence.sequence_code
          _nef_sequence.residue_name
          _nef_sequence.linking
          _nef_sequence.residue_variant
          _nef_sequence.cis_peptide
          _nef_sequence.ccpn_comment
          _nef_sequence.ccpn_chain_role
          _nef_sequence.ccpn_compound_name
          _nef_sequence.ccpn_chain_comment

         1    .   3    HIS   .   .   .   .   .   Sec5   .    

       stop_

    save_

    """
    test_frame = Saveframe.from_string(TEST_DATA)
    chains = frame_to_chains(test_frame)

    assert chains == list([])

CHAIN_STARTS_TEST_DATA = [
        SequenceResidue(chain_code='A', sequence_code=-1, residue_name='ALA'),
        SequenceResidue(chain_code='A', sequence_code=2, residue_name='ALA'),
        SequenceResidue(chain_code='A', sequence_code='z', residue_name='ALA'),

        SequenceResidue(chain_code='B', sequence_code=2, residue_name='ALA'),
        SequenceResidue(chain_code='B', sequence_code=2, residue_name='ALA'),
        SequenceResidue(chain_code='B', sequence_code=3, residue_name='ALA'),

        SequenceResidue(chain_code='c', sequence_code='y', residue_name='ALA')
    ]

def test_count_chains():
    test_frame = Saveframe.from_string(TEST_DATA_MULTI_CHAIN)

    result = {}
    for chain in 'ABC':
        result[chain] = count_residues(test_frame, chain)

    EXPECTED = {'A': {'HIS': 1, 'MET': 1}, 'B': {'ARG': 1, 'GLN': 1}, 'C': {'PRO': 1}}
    assert result == EXPECTED

def test_get_chain_starts():

    EXPECTED = {
        'A': -1,
        'B': 2,
    }

    assert EXPECTED == get_chain_starts(CHAIN_STARTS_TEST_DATA)


def test_offset_chains():

    EXPECTED = [
        SequenceResidue(chain_code='A', sequence_code=1, residue_name='ALA'),
        SequenceResidue(chain_code='A', sequence_code=4, residue_name='ALA'),
        SequenceResidue(chain_code='A', sequence_code='z', residue_name='ALA'),

        SequenceResidue(chain_code='B', sequence_code=0, residue_name='ALA'),
        SequenceResidue(chain_code='B', sequence_code=0, residue_name='ALA'),
        SequenceResidue(chain_code='B', sequence_code=1, residue_name='ALA'),

        SequenceResidue(chain_code='c', sequence_code='y', residue_name='ALA')
    ]

    offsets = {
        'A': 2,
        'B': -2,
        'C': 1,
        'D': 3
    }

    result = offset_chain_residues(CHAIN_STARTS_TEST_DATA, offsets)

    for pair in zip(EXPECTED,result):
        print(pair)

    assert EXPECTED == result


def test_chain_code_iter_basic():
    result = [chain_code for chain_code in islice(chain_code_iter(),3)]

    EXPECTED = list('ABC')

    assert EXPECTED == result


def test_chain_code_iter_with_user():
    result = [chain_code for chain_code in islice(chain_code_iter('D.E'),3)]

    EXPECTED = list('DEA')

    assert EXPECTED == result

def test_chain_code_iter_with_exclude():
    result = [chain_code for chain_code in islice(chain_code_iter(exclude='B.C'),3)]

    EXPECTED = list('ADE')

    assert EXPECTED == result


if __name__ == '__main__':
    pytest.main([f'{__file__}', '-vv'])


import pytest
from typer.testing import CliRunner

from lib.test_lib import assert_lines_match, isolate_frame, path_in_test_data

MOLECULAR_SYSTEM = "nef_molecular_system"
FASTA_IMPORT_SEQUENCE = ["fasta", "import", "sequence"]

runner = CliRunner()


@pytest.fixture
def using_fasta():
    # register the module under test
    import transcoders.fasta  # noqa: F401


EXPECTED_3AA = """\
save_nef_molecular_system
   _nef_molecular_system.sf_category   nef_molecular_system
   _nef_molecular_system.sf_framecode  nef_molecular_system

   loop_
      _nef_sequence.index
      _nef_sequence.chain_code
      _nef_sequence.sequence_code
      _nef_sequence.residue_name
      _nef_sequence.linking
      _nef_sequence.residue_variant
      _nef_sequence.cis_peptide

     1   A   1   ALA   start    .   .
     2   A   2   ALA   middle   .   .
     3   A   3   ALA   end      .   .

   stop_

save_"""


# noinspection PyUnusedLocal
def test_3aa(typer_app, using_fasta, clear_cache, monkeypatch):

    monkeypatch.setattr("sys.stdin.isatty", lambda: False)

    path = path_in_test_data(__file__, "3aa.fasta", local=True)
    result = runner.invoke(typer_app, [*FASTA_IMPORT_SEQUENCE, path])

    print(result.stdout)

    assert result.exit_code == 0

    mol_sys_result = isolate_frame(result.stdout, "%s" % MOLECULAR_SYSTEM)

    assert_lines_match(EXPECTED_3AA, mol_sys_result)


EXPECTED_3A_AB = """\
save_nef_molecular_system
   _nef_molecular_system.sf_category   nef_molecular_system
   _nef_molecular_system.sf_framecode  nef_molecular_system

   loop_
      _nef_sequence.index
      _nef_sequence.chain_code
      _nef_sequence.sequence_code
      _nef_sequence.residue_name
      _nef_sequence.linking
      _nef_sequence.residue_variant
      _nef_sequence.cis_peptide

     1   A    1   ALA   start    .   .
     2   A    2   ALA   middle   .   .
     3   A    3   ALA   end      .   .
     4   B    1   ALA   start    .   .
     5   B    2   ALA   middle   .   .
     6   B    3   ALA   end      .   .

   stop_

save_"""


# noinspection PyUnusedLocal
def test_3aa_x2(typer_app, using_fasta, clear_cache, monkeypatch):

    monkeypatch.setattr("sys.stdin.isatty", lambda: False)

    path = path_in_test_data(__file__, "3aa_x2.fasta")
    result = runner.invoke(typer_app, [*FASTA_IMPORT_SEQUENCE, path])

    assert result.exit_code == 0

    mol_sys_result = isolate_frame(result.stdout, "%s" % MOLECULAR_SYSTEM)

    assert_lines_match(EXPECTED_3A_AB, mol_sys_result)


EXPECTED_3A_AB_B_start_11 = """\
save_nef_molecular_system
   _nef_molecular_system.sf_category   nef_molecular_system
   _nef_molecular_system.sf_framecode  nef_molecular_system

   loop_
      _nef_sequence.index
      _nef_sequence.chain_code
      _nef_sequence.sequence_code
      _nef_sequence.residue_name
      _nef_sequence.linking
      _nef_sequence.residue_variant
      _nef_sequence.cis_peptide

     1   A    1   ALA   start    .   .
     2   A    2   ALA   middle   .   .
     3   A    3   ALA   end      .   .
     4   B   11   ALA   start    .   .
     5   B   12   ALA   middle   .   .
     6   B   13   ALA   end      .   .

   stop_

save_"""


# noinspection PyUnusedLocal
def test_3aa_x2_off_10_b(typer_app, using_fasta, clear_cache, monkeypatch):

    monkeypatch.setattr("sys.stdin.isatty", lambda: False)

    path = path_in_test_data(__file__, "3aa_x2.fasta")
    result = runner.invoke(
        typer_app, [*FASTA_IMPORT_SEQUENCE, "--starts", "1,11", path]
    )

    assert result.exit_code == 0

    mol_sys_result = isolate_frame(result.stdout, "%s" % MOLECULAR_SYSTEM)

    assert_lines_match(EXPECTED_3A_AB_B_start_11, mol_sys_result)



from lib.structures import  ShiftList, ShiftData, AtomLabel
from textwrap import dedent
from transcoders.nmrview.nmrview_lib import  parse_shifts
import pytest
from lib.test_lib import assert_lines_match, isolate_frame, path_in_test_data, run_and_report

LIST_FRAMES = ['frames', 'list']


@pytest.fixture
def using_list():
    # register the module under test
    import tools.frames

# noinspection PyUnusedLocal
def test_frame_basic(typer_app, using_list, monkeypatch):

    monkeypatch.setattr('sys.stdin.isatty', lambda: False)

    path = path_in_test_data(__file__, 'frames.nef')
    result = run_and_report(typer_app, [*LIST_FRAMES, '--in', path])

    if result.exit_code != 0:
        print('INFO: stdout from failed read:\n', result.stdout)

    assert result.exit_code == 0

    EXPECTED =  '''\
        entry test

        nef_nmr_meta_data  nef_molecular_system
    '''

    assert_lines_match (EXPECTED, result.stdout)


# noinspection PyUnusedLocal
def test_frame_file_first(typer_app, using_list, monkeypatch):

    monkeypatch.setattr('sys.stdin.isatty', lambda: False)

    path = path_in_test_data(__file__, 'frames.nef')
    result = run_and_report(typer_app, [*LIST_FRAMES, path])

    if result.exit_code != 0:
        print('INFO: stdout from failed read:\n', result.stdout)

    assert result.exit_code == 0

    EXPECTED =  '''\
        entry test

        nef_nmr_meta_data  nef_molecular_system
    '''

    assert_lines_match (EXPECTED, result.stdout)

# noinspection PyUnusedLocal
def test_frame_file_first_and_select(typer_app, using_list, monkeypatch):

    monkeypatch.setattr('sys.stdin.isatty', lambda: False)

    path = path_in_test_data(__file__, 'frames.nef')
    result = run_and_report(typer_app, [*LIST_FRAMES, path, 'meta_data'])

    print('result',f'-{result.stdout}-')

    EXPECTED =  '''\
        entry test

        nef_nmr_meta_data
    '''

    assert_lines_match (EXPECTED, result.stdout)


import pytest

import lib
from lib.test_lib import (
    assert_lines_match,
    isolate_frame,
    path_in_test_data,
    run_and_report,
)

MOLECULAR_SYSTEM_NMRVIEW = "nef_spectrum_nmrview"
NMRVIEW_IMPORT_PEAKS = ["nmrview", "import", "peaks"]


@pytest.fixture
def using_nmrview():
    # register the module under test
    import transcoders.nmrview  # noqa: F401


EXPECTED_4AA = open(
    path_in_test_data(__file__, "nmrview_expected_4aa.txt", local=True)
).read()


# noinspection PyUnusedLocal
def test_3peaks(typer_app, using_nmrview, clear_cache, monkeypatch):

    # reading stdin doesn't work in pytest so for a clean header
    # TODO move to conftest.py
    monkeypatch.setattr(lib.util, "get_pipe_file", lambda x: None)
    peaks_path = path_in_test_data(__file__, "4peaks.xpk", local=True)
    sequence_path = path_in_test_data(__file__, "4peaks.seq")

    args = [
        *NMRVIEW_IMPORT_PEAKS,
        "--sequence",
        sequence_path,
        "--axis",
        "1H.1H.15N",
        peaks_path,
    ]
    result = run_and_report(typer_app, args)

    result = isolate_frame(result.stdout, "nef_nmr_spectrum_simnoe")

    assert_lines_match(EXPECTED_4AA, result)


def test_3peaks_bad_axis_codes(typer_app, using_nmrview, clear_cache, monkeypatch):

    # reading stdin doesn't work in pytest so for a clean header
    # TODO move to conftest.py
    monkeypatch.setattr(lib.util, "get_pipe_file", lambda x: None)
    peaks_path = path_in_test_data(__file__, "4peaks.xpk")
    sequence_path = path_in_test_data(__file__, "4peaks.seq")
    args = [
        *NMRVIEW_IMPORT_PEAKS,
        "--sequence",
        sequence_path,
        "--axis",
        "1H,15N",
        peaks_path,
    ]
    result = run_and_report(typer_app, args, expected_exit_code=1)

    EXPECTED = """\
    ERROR: can't find isotope code for axis 2 got axis codes 1H,15N
    exiting..."""

    assert_lines_match(EXPECTED, result.stdout)


EXPECTED_JOE_CLIC = """\
    save_nef_nmr_spectrum_trosyrdcjoe
       _nef_nmr_spectrum.sf_category          nef_nmr_spectrum
       _nef_nmr_spectrum.sf_framecode         nef_nmr_spectrum_trosyrdcjoe
       _nef_nmr_spectrum.num_dimensions       2
       _nef_nmr_spectrum.chemical_shift_list  .

       loop_
          _nef_spectrum_dimension.dimension_id
          _nef_spectrum_dimension.axis_unit
          _nef_spectrum_dimension.axis_code
          _nef_spectrum_dimension.spectrometer_frequency
          _nef_spectrum_dimension.spectral_width
          _nef_spectrum_dimension.value_first_point
          _nef_spectrum_dimension.folding
          _nef_spectrum_dimension.absolute_peak_positions
          _nef_spectrum_dimension.is_acquisition



           1   ppm   1H   600.052978516    6.1679  .   circular   true   .
           2   ppm   1H   60.8100013733   35.9997  .   circular   true   .


       stop_

       loop_
          _nef_spectrum_dimension_transfer.dimension_1
          _nef_spectrum_dimension_transfer.dimension_2
          _nef_spectrum_dimension_transfer.transfer_type


       stop_

       loop_
          _nef_peak.index
          _nef_peak.peak_id
          _nef_peak.volume
          _nef_peak.volume_uncertainty
          _nef_peak.height
          _nef_peak.height_uncertainty
          _nef_peak.position_1
          _nef_peak.position_uncertainty_1
          _nef_peak.position_2
          _nef_peak.position_uncertainty_2
          _nef_peak.chain_code_1
          _nef_peak.sequence_code_1
          _nef_peak.residue_name_1
          _nef_peak.atom_name_1
          _nef_peak.chain_code_2
          _nef_peak.sequence_code_2
          _nef_peak.residue_name_2
          _nef_peak.atom_name_2

          1   0   0.254054695368   .   0.0368   .   8.15968   .   126.97411   .   A   2   ala   H   A   2   ala   N

       stop_

    save_
"""


# noinspection PyUnusedLocal
def test_joe_bad_sweep_widths(typer_app, using_nmrview, monkeypatch):

    # reading stdin doesn't work in pytest so for a clean header
    # TODO move to conftest.py
    monkeypatch.setattr(lib.util, "get_pipe_file", lambda x: None)
    peaks_path = path_in_test_data(__file__, "joe_clic.xpk")
    sequence_path = path_in_test_data(__file__, "4peaks.seq")

    command = [
        *NMRVIEW_IMPORT_PEAKS,
        "--sequence",
        sequence_path,
        "--axis",
        "1H.1H.15N",
        peaks_path,
    ]
    result = run_and_report(typer_app, command)

    result = isolate_frame(result.stdout, "nef_nmr_spectrum_trosyrdcjoe")

    assert_lines_match(EXPECTED_JOE_CLIC, result)


if __name__ == "__main__":
    pytest.main([__file__, "-vv"])



# todo sort out appdatalists
# todo derived methods
# todo remove parent links on json dump
# todo unify tree iteerators!

import typing
from enum import auto
from typing import List, Union, Optional, Literal, TypeVar, Callable, Annotated, get_origin, Tuple, get_args

from cachetools import LRUCache
from pydantic import BaseModel, Field
from strenum import StrEnum
from lib.translation.object_iter import ObjectIter, object_children, item_only, return_entity



class PCMeta(BaseModel.__class__):
    _object_map = {}

    def __call__(self, *args, **kwargs):
        result = super(PCMeta, self).__call__(*args, **kwargs)

        if hasattr(result, '__post_init__') and isinstance(result.__post_init__, Callable):
            result.__post_init__()

        return result

class ID(int):
    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v):

        try:
            result = cls(v)
        except (ValueError, TypeError) as e:
            raise ValueError(f'invalid ID format [should look like an integer: {v} error was: {e}],')

        return result

T = TypeVar('T')


class IDTuple(Tuple):
    is_idlist = True
    pass


class IDList(List):
    is_idlist = True
    pass


# these are distinguished by containing IDLists / IDTuples
# these trigger replacement of IDs by instances
# IDs = Union[IDList[ID], ID]
# Types = Union[List[T], T]
# OptionalIDs = Optional[IDs]

# containers of a specific type [child containers] - only some of these are created on their own as indicated
ItemOfType = Annotated[T, None]
PairOfType = Tuple[T, T] # not created standalone
TripletOfType = Tuple[T, T, T] # not created standalone
QuartetOfType = Tuple[T, T, T, T] # not created standalone
OneOrMoreOfType = Union[List[T], T]
OptionalOneOrMoreOfType = Optional[OneOrMoreOfType]

# containers for IDs - not used on their own
ItemOfID = Annotated[ID, None]
PairOfID = IDTuple[ID, ID]
TripletOfID = IDTuple[ID, ID, ID]
QuartetOfID = IDTuple[ID, ID, ID, ID]
OneOrMoreOfID = Union[IDList[ID], ID]
# OptionalOneOrMoreOfID = Optional[OneOrMoreOfID] - no such instances currently

# containers of IDs or types [links]
Item = Union[T, ID]
Pair = Union[PairOfType[T], PairOfID]
Triplet = Union[TripletOfType[T], TripletOfID]
Quartet = Union[QuartetOfType[T], QuartetOfID]
OneOrMore = Union[OneOrMoreOfType[T], OneOrMoreOfID]

OptionalItem = Optional[Item[T]]
OptionalOneOrMore = Optional[OneOrMore[T]]


class Linkage(StrEnum):
    parent =  auto()
    children =  auto()
    link = auto()
    attribute = auto()
    implementation = auto()
    derived_link = auto()
    derived_attribute = auto()
    cross_package_link = auto()

    none = auto()

# attribute that is a parent
Parent = Annotated[Optional[ItemOfType[T]], Linkage.parent] # optional as they are set at runtime

# attribute that is a container of a child or children
Children = Annotated[T, Linkage.children]

# attribute that is a link between objects but not a child
Link = Annotated[T, Linkage.link]

Attribute = Annotated[T, Linkage.attribute]

Implementation = Annotated[T, Linkage.implementation]

DerivedLink = Annotated[Optional[T], Linkage.derived_link] # optional as they are set at runtime

CrossPackageLink = Annotated[Optional[T], Linkage.cross_package_link] # optional as they are set at runtime

class PCBaseModel(BaseModel, metaclass=PCMeta):

    ID: Implementation[Optional[int]] = None  # not required by appdata maybe we need further level of base model

    @classmethod
    def _get_field_annotations(cls, field_name):
        annotations = cls.__annotations__[field_name] if field_name in cls.__annotations__ else []
        if annotations and not isinstance(annotations, list):
            annotations = [annotations]

        return annotations

    @classmethod
    def _field_is_id_list(cls, field_name):
        annotations = cls._get_field_annotations(field_name)


        origin = get_origin(annotations)
        result = hasattr(origin, 'is_idlist')

        if not result:
            args = typing.get_args(annotations)
            arg_origins = [hasattr(get_origin(arg), 'is_idlist') for arg in args if get_origin(arg)]
            result = True in arg_origins

        return result


    def __init__(self, *args, **kwargs):
        super(PCBaseModel, self).__init__(*args, **kwargs)

        if self.ID is not None:
            # noinspection PyProtectedMember,PyUnresolvedReferences
            self.__class__._object_map[self.ID] = self

    def __str__(self) -> str:
        return f'<class {str(self.__class__)[8:-2]} {self.ID}>'

    __repr__ = __str__

# forward references for type system
class ChemAtom(PCBaseModel):
    ...


class ChemBond(PCBaseModel):
    ...


class ChemCompVar(PCBaseModel):
    ...


class ChemTorsion(PCBaseModel):
    ...


class LinkEnd(PCBaseModel):
    ...


class ChemTorsionSysName(PCBaseModel):
    ...


class ChemAtomSet(PCBaseModel):
    ...


class ChemCompSysName(PCBaseModel):
    ...


class NamingSystem(PCBaseModel):
    ...


class ChemAngle(PCBaseModel):
    ...


class AbstractChemAtom(PCBaseModel):
    ...

class ChemComp(PCBaseModel):
    ...

# add UML cardinality comments..

class AtomChirality(StrEnum):
    R = auto()
    S = auto()
    unknown = auto()


class Stereochemistry(PCBaseModel):
    serial: Attribute[int]  # 1..1
    stereoClass: Attribute[str]  # 1..1
    value: Attribute[str]  # 1..1

    chemAtoms: Link[OptionalOneOrMore[AbstractChemAtom]]  # 0..*
    chemComp: Parent[ChemComp] # 1..1
    coreAtoms: Link[OptionalOneOrMore[AbstractChemAtom]]  # 0..*
    parent: Parent[ChemComp] # 1..1                                            # 1..1
    # refStereoChemistry  external to package             # 0..1


class ChemAngle(PCBaseModel):
    chemAtoms: Link[Triplet[ChemAtom]]  # 3..3
    chemComp: Parent[ChemComp]  # 1..1
    parent: Parent[ChemComp]  # 1..1

class AbstractChemAtom(PCBaseModel):
    name: Attribute[str]  # 1..1
    subType: Attribute[int] = 1  # 1..1

    chemAngles: Link[OptionalOneOrMore[ChemAngle]]  # 0..*
    chemBonds: Link[OptionalOneOrMore[ChemBond]]  # 0..*
    chemComp: Parent[ChemComp]  # 1..1
    chemCompVars: Link[OptionalOneOrMore[ChemCompVar]]  # 0..*
    chemTorsions: Link[OptionalOneOrMore[ChemTorsion]]  # 0..*
    coreStereochemistries: Link[OptionalOneOrMore[Stereochemistry]]  # 0..*
    parent: Parent[ChemComp]  # 1..1
    stereochemistries: Link[OptionalOneOrMore[Stereochemistry]]  # 0..*


class ChemAtom(AbstractChemAtom):
    type: Attribute[Literal['ChemAtom']]
    chirality: Attribute[OptionalItem[AtomChirality]]  # 0..1
    elementSymbol: Attribute[str]  # 1..1
    nuclGroupType: Attribute[Optional[str]]  # 0..1
    shortVegaType: Attribute[Optional[str]]  # 0..1
    waterExcangeable: Attribute[bool] = False  # 1..1

    boundLinkEnds: Link[OptionalOneOrMore[LinkEnd]]  # 0..*
    chemAtomSet: Link[OptionalItem[ChemAtomSet]]  # 0..1
    chemComp: Parent[ChemComp]  # 1..1
    # chemElement # derived                               # 1..1
    parent: Parent[ChemComp]  # 1..1
    remotelinkEnd: Link[OptionalOneOrMore[LinkEnd]]  # 0..*


class LinkAtom(AbstractChemAtom):
    type: Literal['LinkAtom']
    elementSymbol: Optional[str]  # 0..1
    boundLinkEnd: Link[OptionalItem[LinkEnd]]  # 0..1
    chemComp: Parent[ChemComp]  # 1..1
    parent: Parent[ChemComp]  # 1..1
    remotelinkEnd: Link[OptionalItem[LinkEnd]]  # 0..1


class ChemTorsion(PCBaseModel):
    name: Attribute[str]  # 0..1

    chemAtoms: Link[Quartet[ChemAtom]]  # 4..4
    chemComp: Parent[ChemComp]  # 1..1
    parent: Parent[ChemComp]  # 1..1
    sysNames: Link[OptionalOneOrMore[ChemTorsionSysName]]  # 0..*


class ChemAtomSet(PCBaseModel):
    distCorr: Attribute[float] = 0.0  # 1..1
    # elementSymbol  #  derived                            # 1..1
    isEquivalent: Attribute[Optional[bool]]  # 0..1
    isProchiral: Attribute[Optional[bool]]  # 1..1  # no default value bu not always present
    name: Attribute[str]  # 1..1
    subType: Attribute[int] = 1  # 1..1

    chemAtomSet: Link[OptionalItem[ChemAtomSet]]  # 0..1
    chemAtomSets: Link[OptionalOneOrMore[ChemAtomSet]]  # 0..*
    chemAtoms: Link[OptionalOneOrMore[ChemAtom]]  # 1..1
    chemComp: Parent[ChemComp]  # 1..1
    parent: Parent[ChemComp]  # 1..1

class BondType(StrEnum):
    single = auto()
    double = auto()
    triple = auto()
    aromatic = auto()
    dative = auto()
    singleplanar = auto()


class BondStereochemistry(StrEnum):
    E = auto()
    Z = auto()
    unknown = auto()


class ChemBond(PCBaseModel):
    bondType: Attribute[BondType] = BondType.single  # 1..1
    stereochem: Attribute[Optional[BondStereochemistry]]  # 0..1

    chemAtoms: Link[Pair[ChemAtom]]  # 2..2
    chemComp: Parent[ChemComp]  # 1..1
    parent: Parent[ChemComp]  # 1..1

class ChemCompLinking(StrEnum):
    start = auto()
    middle = auto()
    end = auto()
    none = auto()
    any = auto()


class ChemCompSysName(PCBaseModel):
    sysName: Attribute[str]  # 1..1

    # chemCompVars # derived                                # 0..*
    namingSystem: Parent[NamingSystem]  # 1..1
    specificChemCompvars: Link[OptionalOneOrMore[ChemCompVar]]  # 0..*
    parent: Parent[NamingSystem]  # 1..1



class ChemCompVar(PCBaseModel):
    descriptor: Attribute[str]  # 1..1
    formalCharge: Attribute[Optional[int]]  # 1..1 but apparently no default...
    # formula # derived str
    glycoCtCode: Attribute[Optional[str]]  # 0..1
    isAromatic: Attribute[bool]  = False  # 1..1 but apparently no default...
    isDefaultVar: Attribute[bool] = False  # 1..1
    isParamagnetic: Attribute[bool] = False  # 1..1 but apparently no default...
    linking: Attribute[ChemCompLinking]  # 1..1
    # molecularMass derived                                   # 1..1
    # name derived                                            # 0..1
    nonStereoSmiles: Attribute[Optional[str]]  # 0..1
    stereoSmiles: Attribute[Optional[str]]  # 0..1
    varName: Attribute[Optional[str]]  # 0..1

    # chemAngles    derived                                   # 0..*
    # chemAtomSets  derived                                   # 0..*
    chemAtoms: Link[OptionalOneOrMore[AbstractChemAtom]]  # 0..*
    # chemBonds     derived                                   # 0..*
    chemComp: Parent[ChemComp]  # 1..1
    # chemCompSysNames derived                                # 0..*
    # chemTorsions derived                                    # 0..*
    # linkEnds derived                                        # 0..*
    parent: Parent[ChemComp]  # 1..1
    specificSysNames: Link[OptionalOneOrMore[ChemCompSysName]]  # 0..*


class LinkEnd(PCBaseModel):
    linkCode: Attribute[str]  # should be enum                         # 1..1

    boundChemAtom: Link[OptionalItem[ChemAtom]]  # 1..1 is required but not always present?
    boundLinkAtom: Link[OptionalItem[LinkAtom]]  # 1..1 is required but not always present?
    chemComp: Parent[ChemComp]  # 1..1
    remoteChemAtom: Link[OptionalItem[ChemAtom]]  # 1..1 is required but not always present?
    remoteLinkAtom: Link[OptionalItem[LinkAtom]]  # 1..1 is required but not always present?
    parent: Parent[ChemComp]  # 1..1


class AtomSysName(PCBaseModel):
    altSysName: Attribute[OptionalOneOrMore[str]]  # 0..*
    atomName: Attribute[str]  # 1..1
    atomSubType: Attribute[int] = 1  # 1..1
    sysName: Attribute[Optional[str]]  # 1..1 is required but not always present?

    namingSystem: Parent[NamingSystem]  # 1..1
    parent: Parent[NamingSystem]  # 1..1


class ChemTorsionSysName(PCBaseModel):
    sysName: Attribute[Optional[str]] # 1..1
    chemTorsion: Link[OptionalItem[ChemTorsion]]  # 1..1 is required but not always present?
    namingSystem: Parent[NamingSystem]  # 1..1
    parent: Parent[NamingSystem]  # 1..1


class NamingSystem(PCBaseModel):
    name: Attribute[str]  # 1..1

    atomReference: Link[OptionalItem[NamingSystem]]  # 0..1
    atomSetReference: Link[OptionalItem[NamingSystem]]  # 0..1
    atomSetVariantSystems: Link[OptionalOneOrMore[NamingSystem]]  # 0..*
    atomSysNames: Children[OptionalOneOrMore[AtomSysName]]  # 0..*
    atomVariantSystems: Link[OptionalOneOrMore[NamingSystem]]  # 0..*
    chemComp: Parent[ChemComp]  # 1..1
    chemCompSysNames: Children[OptionalOneOrMore[ChemCompSysName]]  # 0..*
    chemTorsionSysNames: Children[OptionalOneOrMore[ChemTorsionSysName]]  # 0..*
    # mainChemCompSysName # derived                            # 0..1
    parent: Parent[ChemComp]  # 1..1


class ApplicationData(PCBaseModel):
    application: Attribute[str]  # 1..1
    keyword: Attribute[str]  # 1..1

    def __str__(self):
        return f'{self.__class__} {self.application}:{self.keyword}'

    __repr__ = __str__


class AppDataStr(ApplicationData):
    type: Literal['AppDataString']
    value: Attribute[OneOrMore[str]]  # 1..1 we are getting lists of strings here where it should be lists of AppDataString...


class AppDataInt(ApplicationData):
    type: Literal['ApplicationDataInt']
    value: Attribute[OneOrMore[int]]  # 1..1 we are getting lists of ints here where it should be lists of AppDataString...


class AppDataBoolean(ApplicationData):
    type: Literal['AppDataBoolean']
    value: Attribute[bool]  # 1..1 seems to be correct here...


AppDataTypes = Union[AppDataStr, AppDataInt, AppDataBoolean]


def object_children_root_only(target, parent):

    linkage = get_linkage(parent.target, parent.selector)
    elems = object_children(target, parent)

    if isinstance(target, (list, PCBaseModel)):
        if linkage in (Linkage.link, Linkage.parent):
            for elem in elems:
                elem.is_leaf = True

    return elems

class MolType(StrEnum):
    protein = auto()
    DNA = auto()
    RNA = auto()
    carbohydrate = auto()
    other = auto()

# inspect.get_annotations only available in 3.10
class Annotations_Cache:

    def __init__(self, size):
        self._cache = LRUCache(size)


    def get_annotations(self, o):


        object_id = (o.__class__.__name__, o.__class__.__module__, id(o))

        if object_id in self._cache:
             result = self._cache[object_id]

        else:
            if isinstance(o, type):
                attr_dict = getattr(o, '__dict__', {})
                result = attr_dict.get('__annotations__', {})
            else:
                result = getattr(o, '__annotations__', {})

            if result is None:
                result = {}

            self._cache[object_id] = result

        return result

# inspect.get_annotations recreates the data everytime so cache it !!!
annotation_cache = Annotations_Cache(30)

def get_annotations(o: object, include_super_classes:bool = True) -> typing.Dict[str, type]:

    annotations = annotation_cache.get_annotations(o)

    if include_super_classes:
        for super_class in o.__class__.mro():
            super_class_annotations = annotation_cache.get_annotations(super_class)

            super_class_annotations.update(annotations)
            annotations =  super_class_annotations

    return annotations


def get_item_linkage(item: ObjectIter.Entry) -> Linkage:
    return get_linkage(item.target, item.selector)


def get_linkage(target: PCBaseModel, selector: str) -> Linkage:

    annotations = get_annotations(target)

    result = Linkage.none
    if selector and selector in annotations:
        annotation = annotations[selector]
        if (get_origin(annotation) is typing.Annotated):
            args = set(get_args(annotation))

            link_type = args.intersection(Linkage.__members__.keys())
            result = next(iter(link_type)) if len(link_type) else Linkage.none

    return result


class ChemComp(PCBaseModel):



    baseGlycoCode: Attribute[Optional[str]]  # 0..1
    beilsteinCode: Attribute[Optional[str]]  # 0..1
    caseRegCode: Attribute[Optional[str]]  # 0..1
    ccpCode: Attribute[str]  # 1..1
    code1Letter: Attribute[Optional[str]]  # 0..1
    code3Letter: Attribute[Optional[str]]  # 0..1
    commonNames: Attribute[OptionalOneOrMore[str]]  # 0..*
    details: Attribute[Optional[str]]  # 0..1
    hasStdChirality: Attribute[Optional[bool]]  # 0..1
    # isLinearPolymer: bool   derived                        # 1..1
    keywords: Attribute[OptionalOneOrMore[str]] # 0..*
    merckCode: Attribute[Optional[str]]  # 0..1
    molType: Attribute[MolType]  # 1..1
    name: Attribute[Optional[str]]  # 0..1
    sigmaAldrichCode: Attribute[Optional[str]]  # 0..1
    stdChemCompCode: Attribute[Optional[str]]  # 0..1

    # should be in parent class...
    guid: Attribute[str]
    createdBy: Attribute[str]

    chemAngles: Children[OptionalOneOrMoreOfType[ChemAngle]]  # 0..*
    chemAtomSets:  Children[OptionalOneOrMoreOfType[ChemAtomSet]]  # 0..*
    chemAtoms:  Children[List[Annotated[Union[ChemAtom, LinkAtom], Field(discriminator='type')]]]  # 0..*
    chemBonds:  Children[OptionalOneOrMoreOfType[ChemBond]]  # 0..*
    chemCompVars: Children[OptionalOneOrMoreOfType[ChemCompVar]] # 0..*
    chemTorsions:  Children[OptionalOneOrMoreOfType[ChemTorsion]]  # 0..*
    memopsRoot: Parent[object] # Parent[MemopsRoot] parent - not implemented # 1..1
    linkEnds: Children[OptionalOneOrMoreOfType[LinkEnd]]  # 0..*
    namingSystems: Children[OptionalOneOrMoreOfType[NamingSystem]]  # 0..*
    parent: Parent[object] # : Parent[MemopsRoot]  # 1..1
    # residueTypeProbabilities - out of package not implemented                            # 0..*
    # stdChemComp  #derived                                                                # 0..1
    stereochemistries:  Children[OptionalOneOrMoreOfType[Stereochemistry]]  # 0..*

    applicationData:  Children[OptionalOneOrMore[Annotated[AppDataTypes, Field(discriminator='type')]]]

    def __post_init__(self):
        self._make_links()

    def _make_links(self):
        object_iter = ObjectIter(result_generator=item_only)
        for_parent = []
        for_children = []

        for i, item in enumerate(object_iter.iter(self)):
            linkage = get_item_linkage(item)

            if linkage == Linkage.parent:
                for_parent.append(item)
            elif linkage == Linkage.link:
                if item.value is not None:
                    for_children.append(item)

        for item in for_parent:
            setattr(item.target, item.selector, object_iter.get_parent(object_iter.get_parent(item)).target)

        for j, item in enumerate(for_children):
            value = item.value
            if isinstance(value, ID):

                setattr(item.target, item.selector, self.__class__._object_map[item.value])
            elif isinstance(value, tuple):

                new_value = [None] * len(value)
                for i, elem in enumerate(item.value):
                    if isinstance(elem, ID):
                        new_value[i] = self.__class__._object_map[elem]
                setattr(item.target, item.selector, tuple(new_value))
            elif isinstance(value, list):

                for i, elem in enumerate(item.value):
                    if isinstance(elem, ID):
                        value[i] = self.__class__._object_map[elem]

    def json(self, original=False):

        if original:
            result = super(ChemComp, self).json()
        else:
            copy_of_data = self.copy(deep=True)
            copy_of_data.unlink()
            result = copy_of_data.json(original=True)
        return result


    def unlink(self):
        object_iter = ObjectIter(result_generator=item_only, child_lister=object_children_root_only)
        for_parent = []
        for_children = []
        for i, item in enumerate(object_iter.iter(self)):

            linkage = get_item_linkage(item)

            if linkage == Linkage.parent:
                for_parent.append(item)
            elif linkage == Linkage.link:
                if item.value is not None:
                    for_children.append(item)

        for item in for_parent:
            setattr(item.target, item.selector, None)
        for j, item in enumerate(for_children):

            value = item.value
            if isinstance(value, PCBaseModel):

                setattr(item.target, item.selector, value.ID)
            elif isinstance(value, tuple):

                new_value = [None] * len(value)
                for i, elem in enumerate(item.value):

                    if isinstance(elem, PCBaseModel):
                        new_value[i] = elem.ID
                setattr(item.target, item.selector, tuple(new_value))

            elif isinstance(value, list):

                for i, elem in enumerate(item.value):
                    if isinstance(elem, PCBaseModel):
                        value[i] = elem.ID



from lib.test_lib import assert_lines_match
from transcoders.nmrpipe.nmrpipe_lib import print_pipe_sequence


def test_print_pipe_sequence(capsys):
    TEST_DATA = ['A'] * 110
    EXPECTED = """\
        DATA SEQUENCE AAAAAAAAAA AAAAAAAAAA AAAAAAAAAA AAAAAAAAAA AAAAAAAAAA AAAAAAAAAA AAAAAAAAAA AAAAAAAAAA AAAAAAAAAA AAAAAAAAAA
        DATA SEQUENCE AAAAAAAAAA
    """

    print_pipe_sequence(TEST_DATA)

    assert_lines_match(capsys.readouterr().out, EXPECTED)


#!/usr/bin/env python3
import argparse
import string

from tabulate import tabulate
import sys
import pathlib
import re

from pynmrstar import Entry, Saveframe, Loop


EXIT_ERROR = 1
UNUSED ='.'
filename_matcher=re.compile('fold_([0-9]+).sa.viols')
parser = None
range_split = re.compile('\.\.').split

def exit_error(msg):

        print(f'ERROR: {msg}')
        print(' exiting...')
        sys.exit(EXIT_ERROR)

def read_args():
    global parser
    parser = argparse.ArgumentParser(description='convert xplor nOe violations to NEF')

    parser.add_argument(metavar='<XPLOR>.viols', action='extend', nargs='+', dest='files')

    parser.add_argument('-o', '--output-file', dest='output_file', default=None)

    return parser.parse_args()

def collapse_names(names, depth=2):

    if len(names) == 1:
        result=names[0]
    else:
        digit_by_depth = {}
        for target in names:
            for i in range(1, depth+1):

                if target[-i].isdigit():
                    digit_by_depth.setdefault(-i,set()).add(target[-i])

                else:
                    break

        name_list = list(list(names)[0])
        for depth, unique_digits in digit_by_depth.items():
            if len(unique_digits) > 1:
                name_list[depth] = '%'
        result = ''.join(name_list)

        result = collapse_right_repeated_string(result)
    return result


def collapse_right_repeated_string(target, repeating='%'):

    orig_length = len(target)

    result = target.rstrip(repeating)

    if len(result) < orig_length:
        result = result + repeating

    return result


def viol_to_nef(file_handle, model, file_path, args):
    in_data = False
    results = {}

    for line in file_handle:

        if 'NOE restraints in potential term' in line:
            restraint_number = None
            sub_restraint_id = 0
            pair_number = 0
            current_selections = None
            current_fields = None
            index = 0

            frame_name = line.split(':')[1].split('(')[0].strip()
            in_data = True
            lines = {}
            results[frame_name] = lines

            while not line.startswith('-----'):
                line = next(file_handle)
            line = next(file_handle)

        if in_data and line.startswith('number of restraints'):
            in_data = False
            restraint_number = None
            pair_number = None
            index = 0

        if in_data and len(line.strip()) == 0:
            continue

        if in_data and '-- OR --' in line:
            sub_restraint_id += 1
            pair_number = 0
            continue

        if in_data:
            line = line.strip()

            line = _remove_violated_mark_if_present(line)

            new_restraint_number = _get_id(line, restraint_number)

            pair_number = 0 if new_restraint_number != restraint_number else pair_number
            current_pair_number = pair_number
            pair_number = pair_number+1

            sub_restraint_id = 0 if new_restraint_number != restraint_number else sub_restraint_id
            restraint_number = new_restraint_number
            line = _remove_id(line)

            selections = _get_selections(line, 2, current_selections)
            current_selections = selections
            line = _remove_selections(line, 2)

            line = ' '.join(line.split('!'))
            fields = line.split()
            if not fields:
                fields = current_fields
            current_fields = fields

            calculated = float(fields[0])
            min_bound = float(range_split(fields[1])[0])
            max_bound = float(range_split(fields[1])[1])

            violation = 0.0
            if calculated < min_bound:
                violation = calculated - min_bound
            elif calculated > max_bound:
                violation = calculated - max_bound


            key = (index, model, new_restraint_number, sub_restraint_id, current_pair_number)
            comment = fields[3]
            restraint_list = comment.rstrip(string.digits)
            restraint_identifier = comment[len(restraint_list):]
            result = {'selection-1': selections[0],
                      'selection-2': selections[1],
                      'probability': '.',
                      'calc': calculated,
                      'min': min_bound,
                      'max': max_bound,
                      'dist': calculated,
                      'viol': violation,
                      'restraint-list': restraint_list,
                      'restraint-number': restraint_identifier,
                      'comment': fields[3],
                      'violation-file': file_path.parts[-1],
                      'structure-file': f'{file_path.stem}.cif'

            }

            lines[key] = result
            index += 1


            # print(line.strip().split([')','(']))
            # while((fields:=line.strip().partition(')'))[2:] !=('','')):
            #     line=fields[-1]
            #     print(fields)
            # return
    args.next_index = index
    return results

def _get_id(line, current_id):
    putative_id = line.split()[0].strip()

    result = current_id
    if not putative_id.startswith('('):
        result = int(putative_id)

    return result


def _remove_violated_mark_if_present(line):

    result = line.strip().lstrip('*').strip()
    return result


def _remove_id(line):
    return line.lstrip('0123456789').strip()


def _line_active(line):
    fields = line.split()
    return fields[0] != '*'


def _get_selections(line, count, current_selections):

    results = []
    line = line.strip()
    for i in range(count):
        value, _, line = line.partition(')')
        results.append(value)
        line = line.strip()

    results = [selection.lstrip('(') for selection in results]
    selections = []
    for i, result in enumerate(results):
        selection = []

        if len(result.strip()):
            selections.append(selection)
            segid = result[:4]
            selection.append(segid)
            result = result[4:]
            selection.extend(result.split())
        else:
            selections.append(current_selections[i])

    return selections


def _remove_selections(line,count):

    line = line.strip()
    for i in range(count):
        value, _, line = line.partition(')')
        line = line.strip()

    return line


def tabulate_data(data):
    for restraint_table in data:
        print(restraint_table)
        print('-' * len(restraint_table))
        print()



        table = []
        non_index_headings = ['probability', 'calc', 'min',
                    'max', 'dist', 'viol', 'restraint-list', 'restraint-number', 'comment']
        selections = ['selection-1', 'selection-2' ]

        for indices, datum in data[restraint_table].items():

            line = []
            table.append(line)

            have_path_ids = len(indices) == 5
            if have_path_ids:
                (index, model, id, sub_id,path_id) = indices
            else:
                (index, model, id, sub_id) = indices

            line.append(index+1)
            line.append(model)
            line.append(int(id)+1)
            line.append(sub_id+1)
            if have_path_ids:
                line.append(path_id + 1)
            line.append('.'.join(datum['selection-1']))
            line.append('.'.join(datum['selection-2']))

            for elem in non_index_headings:
                line.append(datum[elem])

        if have_path_ids:
            headings = ['index', 'model', 'id', 'sub id', 'path_id', *selections, *non_index_headings]
        else:
            headings = ['index', 'model', 'id', 'sub id', *selections, *non_index_headings]

        print(tabulate(table, headers=headings))
        print()


def _collapse_pairs(nef_entries):

    result = {}
    for entry_name, data in nef_entries.items():

        pair_selections = {}
        unique_entry_data = {}

        for i,(key, entry) in enumerate(data.items()):

            new_key = key[1:-1]

            pair_selections.setdefault(new_key, []).append((entry['selection-1'],entry['selection-2']))
            unique_entry_data[new_key] = entry

        new_entry = {}
        for i, (new_key, selection_data) in enumerate(pair_selections.items()):

            atoms_1 = set()
            atoms_2 = set()
            for selection_1, selection_2 in selection_data:
                atoms_1.add(selection_1[-1])
                atoms_2.add(selection_2[-1])

            selection_1 = selection_data[0][0]
            selection_2 = selection_data[0][1]

            nef_atom_name_1 = collapse_names(atoms_1)
            nef_atom_name_2 = collapse_names(atoms_2)

            selection_1[-1] = nef_atom_name_1
            selection_2[-1] = nef_atom_name_2

            unique_entry_data[new_key]['selection-1'] = selection_1
            unique_entry_data[new_key]['selection-2'] = selection_2



            new_entry[new_key] = unique_entry_data[new_key]

        result[entry_name] = new_entry


    # ic(result)
    return result


def data_as_nef(overall_result):

    entry = Entry.from_scratch('default')

    for table_name, table_data in overall_result.items():

        category = "ccpn_distance_restraint_violation_list"
        frame_code = f'{category}_{table_name}'
        frame = Saveframe.from_scratch(frame_code, category)
        entry.add_saveframe(frame)

        frame.add_tag("sf_category", category)
        frame.add_tag("sf_framecode", frame_code)
        frame.add_tag("nef_spectrum", f"nef_nmr_spectrum_{list(table_data.values())[0]['restraint-list']}")
        frame.add_tag("nef_restraint_list", f"nef_distance_restraint_list_{list(table_data.values())[0]['restraint-list']}")
        frame.add_tag("program", 'Xplor-NIH')
        frame.add_tag("program_version", UNUSED)
        frame.add_tag("protocol", 'marvin/pasd,refine')
        frame.add_tag("protocol_version", UNUSED)
        frame.add_tag("protocol_parameters", UNUSED)



        lp = Loop.from_scratch()

        tags = ('index', 'model_id', 'restraint_id', 'restraint_sub_id',
                'chain_code_1','sequence_code_1', 'residue_name_1', 'atom_name_1',
                'chain_code_2', 'sequence_code_2', 'residue_name_2', 'atom_name_2',
                'weight', 'probability', 'lower_limit', 'upper_limit', 'distance', 'violation',
                'violation_file', 'structure_file', 'structure_index','nef_peak_id', 'comment'
                )

        lp.set_category('ccpn_restraint_violation')
        lp.add_tag(tags)

        for index, (indices, line_data) in enumerate(table_data.items()):


            indices = list(indices)
            indices = [index,*indices]

            indices[0] += 1
            indices[2] += 1
            indices[3] += 1

            #TODO: conversion of SEGID to chain ID maybe too crude
            selection_1 = line_data['selection-1']
            selection_1[0] = selection_1[0].strip()
            selection_2 = line_data['selection-2']
            selection_2[0] = selection_2[0].strip()

            data = [*indices, *selection_1, *selection_2,
                    1.0, line_data['probability'], line_data['min'], line_data['max'],
                    # GST: this removes trailing rounding errors without loss of accuracy
                    round(line_data['dist'], 10),
                    round(line_data['viol'],10),
                    line_data['violation-file'],line_data['structure-file'], 1,
                    line_data['restraint-number'], line_data['comment']
                    ]

            lp.add_data(data)

        frame.add_loop(lp)

    return str(entry)


if __name__ == '__main__':

    args = read_args()

    overall_result = {}

    for name in args.files:
        path = pathlib.Path(name)

        try:
            model_number = int(filename_matcher.match(path.parts[-1]).group(1))
        except:
            exit_error(f"Couldn't find a model number in {path.parts[-1]} using matcher {filename_matcher.pattern}")

        if not path.is_file():
            exit_error(f'I need an input file, path was {path}')

        with open(path) as fh:
            entries = viol_to_nef(fh, model_number, path, args)

            entries = _collapse_pairs(entries)

            for entry_name, entry in entries.items():
                overall_result.setdefault(entry_name, {}).update(entry)


    #tabulate_data(overall_result)

    result = data_as_nef(overall_result)
    if args.output_file:
        with open(args.output_file,'w') as fh:
            fh.write(result)
    else:
        print(result)

